#include "dtw.h"
#include "common.h"
#include <stdio.h>

  /*********************************************************************
   * FUNCTION: DTWDistance
   * PURPOSE: Dynamic Time Warping Distance
   * INPUTS: 
   *    ppfR[0..swLenR-1, 0..swSize-1] -        reference param matrix
   *    ppfT[0..swLenT-1, 0..swSize-1] -        test param matrix
   *    swLenR                         -        ref. length
   *    swLenT                         -        test length
   *    swSize                         -        par. vec. size
   *    pswPath - vector allocated for path, if NULL, the path won't be
   *              written
   * OUTPUTS: 
   *    pswPath[0..something] ..... - not really well written, do not use. 
   * RETURN: 
   *    DTW-distance (float)   
   * DESCRIPTION:
   *    based on Matlab function of Pascale Jardin & Gen. Baudoin (ESIEE)
   * KEYWORDS:
   ***********************************************************************/
float DTWDistance (float **ppfR, float **ppfT, short swLenR, short swLenT, 
                   short swSize, short *pswPath)
{
   float **g,d,g1,g2,g3;
   short i,j,**trace, ir,it;
   
   g=Alloc_2_Float (swLenT+1, swLenR+1);     /* cumul dist */
   trace=Alloc_2_Short (swLenT+1, swLenR+1); /* for path backtrace */
   /* fill the 1st row and col. with big values, except for [0,0] */
   for (i=1; i<=swLenR; i++) g[0][i]=BIG_FLOAT_VALUE;
   for (i=1; i<=swLenT; i++) g[i][0]=BIG_FLOAT_VALUE;
   g[0][0]=0.0;
   /* now go and make the cumul dist */
   for (i=1; i<=swLenT; i++) {
      for (j=1; j<=swLenR; j++)  {
         d=SqEuclidDistance(ppfT[i-1], ppfR[j-1], swSize);
         g2=g[i-1][j] + d;
         g3=g[i][j-1] + d;
         g1=g[i-1][j-1] + d + d; /* faster than 2*d */
         if (g1 < g2 && g1 < g3) {
           g[i][j] = g1;
           trace[i][j] = 1;
         }
         else
           if (g2 < g3) {
              g[i][j] = g2;
              trace[i][j] = 2;
           }
           else {
              g[i][j] = g3;
              trace[i][j] = 3;
           }
      }
   }
#ifdef DEBUG_DTW
   for(i=0; i<=swLenT; i++) pfv ("", g[i], swLenR+1); pnl();
   for(i=0; i<=swLenT; i++) pswv ("", trace[i], swLenR+1); pnl();
#endif      
   
   /* --------------------------------------------- */ 
   /* if the path requested, going to backtrace it  */
   /* --------------------------------------------- */

   if (pswPath != NULL ) {
     ir=swLenR; it=swLenT;    /* init for backtrace */
     pswPath[ir-1] = it-1;    /* last vector always aligned with last */
 
     while (it + ir > 2) {   /* last point: it=1 and ir=1 */
       /* --- depending on the trace, modif. the pointers --- */
       switch (trace[it][ir]) {
       case 1: it--; ir--; break;
       case 2: it--;       break;
       case 3:       ir--; break;
       }
       /* --- and record to Path --- */
       pswPath[ir-1] = it-1;
     } /* of while */
     
     /* --- the last vector should be always aligned with the last, so
            impose it --- */
     pswPath[swLenR-1] = swLenT-1;

   } /* of path creation */

   d=g[swLenT][swLenR] / (float)(swLenT + swLenR);
   Free_2_Float (g, swLenT+1);
   Free_2_Short (trace, swLenT+1); 
   return (d);
}

  /*********************************************************************
   * FUNCTION: SqEuclidDistance
   * PURPOSE: squqred Euclidian Distance
   * INPUTS: 
   *    pfA[0..swLen-1]      1st vector
   *    pfB[0..swLen-1]      2nd vector
   *    swLen                their length
   * OUTPUTS: 
   *    none
   * RETURN: 
   *    squarred distance (float)   
   * DESCRIPTION:
   * KEYWORDS:
   ***********************************************************************/
float SqEuclidDistance (float *pfA, float *pfB, short swLen)
{
   short i;
   float *f,*g,aux, sum;
   
   sum=0.0; f=pfA; g=pfB;
   for (i=0; i< swLen; i++)  {
      aux = (*f++) - (*g++);
      sum += aux * aux;
   }
   return (sum);
}
